@@ -7,6 +7,7 @@ import signal
 import threading
 import traceback
 from contextlib import contextmanager
+from dataclasses import dataclass
 from datetime import datetime
 from enum import Enum
 from itertools import islice
@@ -193,6 +194,12 @@ class DataSourcesReader:
         self._retry_timeout = value
 
 
+class DataStreamTimeoutBreakStatus(Exception):
+    def __init__(self, break_status):
+        super(DataStreamTimeoutBreakStatus, self).__init__()
+        self.break_status = break_status
+
+
 class NumerousClient:
 
     _execution_id: str
@@ -1254,6 +1261,16 @@ class NumerousClient:
         log.debug('Closing channel')
         self.channel.close()
 
+    def set_data_stream_status(self, scenario, execution, status):
+        self.set_timeseries_custom_meta_data(
+            scenario=scenario, execution=execution, key='status', meta=json.dumps(status)
+        )
+
+    def get_data_stream_status(self, scenario, execution):
+        return json.loads(
+            self.get_timeseries_custom_meta_data(scenario=scenario, execution=execution, key='status')
+        )
+
     def open_write_data_stream(
             self,
             scenario: Optional[str] = None,
@@ -1313,22 +1330,61 @@ class NumerousClient:
                     tags: List[str],
                     start: Optional[float] = None,
                     end: Optional[float] = None,
-                    length: Optional[int] = None
+                    length: Optional[int] = None,
+                    timeout_wait: float = 0.0,
+                    timeout_sleep_for: float = config.DATA_STREAM_DEFAULT_TIMEOUT_SLEEP_FOR,
+                    timeout_break_statuses: Optional[List[Any]] = None,
             ):
-                response = stream.send(
-                    spm_pb2.ReadDataStreamRequest(
-                        scenario=scenario,
-                        execution=execution,
-                        tags=tags,
-                        start=start,
-                        end=end,
-                        length=length,
+                deadline_time = time() + timeout_wait
+                while True:
+                    response = stream.send(
+                        spm_pb2.ReadDataStreamRequest(
+                            scenario=scenario,
+                            execution=execution,
+                            tags=tags,
+                            start=start,
+                            end=end,
+                            length=length,
+                        )
                     )
-                )
-                return response.index, {tag: stream_data.values for tag, stream_data in response.data.items()}
+
+                    if response.index or time() >= deadline_time:
+                        return response.index, {tag: stream_data.values for tag, stream_data in response.data.items()}
+
+                    if timeout_break_statuses:
+                        status = self.get_data_stream_status(scenario, execution)
+                        if status in timeout_break_statuses:
+                            raise DataStreamTimeoutBreakStatus(status)
+
+                    sleep(timeout_sleep_for)
 
         return _ReadDataStream(self)
 
+    @dataclass
+    class DataStreamStats:
+        min: Optional[float]
+        max: Optional[float]
+        len: Optional[int]
+
+    def get_data_stream_stats(
+            self,
+            scenario: Optional[str] = None,
+            execution: Optional[str] = None,
+            tags: Optional[List[str]] = None,
+    ) -> Dict[str, DataStreamStats]:
+        if scenario is None:
+            scenario = self._scenario
+
+        if execution is None:
+            execution = self._execution_id
+
+        request = spm_pb2.GetDataStreamStatsRequest(scenario=scenario, execution=execution, tags=tags)
+        response: spm_pb2.GetDataStreamStatsResponse = self._base_client.GetDataStreamStats(request)
+        return {
+            tag: self.DataStreamStats(min=tag_stats.min, max=tag_stats.max, len=tag_stats.len)
+            for tag, tag_stats in response.stats.items()
+        }
+
 
 @contextmanager
 def open_client(job_id: Optional[str] = None, project: Optional[str] = None, scenario: Optional[str] = None,
